#ifndef SD_L0_H
#define SD_L0_H
#include <Arduino.h>
#include <avr/io.h>
/**
 * SD Chip Select pin
 *
 * Warning if this pin is redefined the hardware SS will pin will be enabled
 * as an output by init().  An avr processor will not function as an SPI
 * master unless SS is set to output mode.
 */
/** The default chip select pin for the SD card is SS. */

#if defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
// Mega
uint8_t const SD_L0_CHIP_SELECT_PIN_DEFAULT = 4;

//------------------------------------------------------------------------------
#elif defined(__AVR_ATmega644P__) || defined(__AVR_ATmega644__)
// SPI port
uint8_t const SD_L0_CHIP_SELECT_PIN_DEFAULT = 4;

//------------------------------------------------------------------------------
#elif defined(__AVR_ATmega32U4__)
// Teensy 2.0
uint8_t const SD_L0_CHIP_SELECT_PIN_DEFAULT = 0;

//------------------------------------------------------------------------------
#elif defined(__AVR_AT90USB646__) || defined(__AVR_AT90USB1286__)
// Teensy++ 1.0 & 2.0
uint8_t const SD_L0_CHIP_SELECT_PIN_DEFAULT = 20;

//------------------------------------------------------------------------------
#else  
// 168 and 328 Arduinos
uint8_t const SD_L0_CHIP_SELECT_PIN_DEFAULT = 4;

#endif  

/**
 * Set chip-select from sd-card to low.
 * Can be a makro or a function.
 */
#define SD_L0_SetCSLow()	{ digitalWrite(SD_L0_CSPin, LOW); }

/**
 * Set chip-select from sd-card to high.
 * Can be a makro or a function.
 */
#define SD_L0_SetCSHigh()	{ digitalWrite(SD_L0_CSPin, HIGH); }

/**
 * Get timestamp. Put your favourite function
 * here to get a timestamp. Timestamp should
 * count not faster than 500us.
 * Return value must be uint16_t! 
 * Overflows doesn't matter as long as 
 * overflow_period is bigger than largest
 * timeout.
 */
#define SD_L0_GetTimestamp()	((uint16_t)millis())

/** 
 * Define timeouts for different functions here
 * based on your GetTimestamp value.
 * All timeouts values should smaller than 30000.
 */
/** init timeout typ. 2000 ms */
#define SD_INIT_TIMEOUT 2000
/** erase timeout typ. 10000 ms */
#define SD_ERASE_TIMEOUT 10000
/** read timeout typ. 300 ms */
#define SD_READ_TIMEOUT 300
/** write time out typ. 600 ms */
#define SD_WRITE_TIMEOUT 600 
/** command timeout typ. 300 ms */
#define SD_COMMAND_TIMEOUT 300

void SD_L0_Init(void);
void SD_L0_DeInit(void);
void SD_L0_SpiSetHighSpeed(void);
uint8_t SD_L0_SpiRecvByte();
void SD_L0_SpiSendByte(uint8_t b);
void SD_L0_SpiRecvBlock(uint8_t *buf, uint16_t nbyte);
void SD_L0_SpiSendBlock(uint8_t token, const uint8_t *buf);

extern uint8_t SD_L0_CSPin;

#endif




//##########################################################################################################################################################



#ifndef SD_L1_H
#define SD_L1_H

#define SD_ENABLE_WRITE_ACCESS          0
#define SD_ENABLE_MULTIBLOCK_ACCESS     1
#define SD_ENABLE_L1_INFORMATIVE        0

// init/deinit functions
// ***************************************
uint8_t     SD_L1_Init();
void        SD_L1_DeInit();

// read access to 512 byte blocks
// ***************************************
uint8_t     SD_L1_ReadBlock(uint32_t blockNumber, uint8_t *dst);
#if SD_ENABLE_MULTIBLOCK_ACCESS
  uint8_t   SD_L1_ReadMBStart(uint32_t blockNumber); 
  uint8_t   SD_L1_ReadMB(uint8_t *dst); 
  uint8_t   SD_L1_ReadMBStop();
#endif /* SD_ENABLE_MULTIBLOCK_ACCESS */

// write access to 512 byte blocks
// ***************************************
#if SD_ENABLE_WRITE_ACCESS
  uint8_t   SD_L1_WriteBlock(uint32_t blockNumber, const uint8_t *src);
  #if SD_ENABLE_MULTIBLOCK_ACCESS
    uint8_t SD_L1_WriteMBStart(uint32_t blockNumber, uint32_t eraseCount);
    uint8_t SD_L1_WriteMB(const uint8_t* src);
    uint8_t SD_L1_WriteMBStop();
  #endif /* SD_ENABLE_MULTIBLOCK_ACCESS */
#endif /* SD_ENABLE_WRITE_ACCESS */


// other functions, just informative
// ***************************************
uint8_t     SD_L1_GetCardType();
#if SD_ENABLE_L1_INFORMATIVE
    uint32_t    SD_L1_GetCardSize();
    uint8_t     SD_L1_ReadCSD(uint8_t *dst);
    uint8_t     SD_L1_ReadCID(uint8_t *dst);
#endif


// error codes
/** timeout error for command CMD0 (initialize card in SPI mode), signal problem */
#define SD_CARD_ERROR_CMD0          0x01
/** CMD8 was not accepted - not a valid SD card */
#define SD_CARD_ERROR_CMD8          0x02
/** ACMD41 initialization process timeout */
#define SD_CARD_ERROR_ACMD41        0x03
/** card returned an error response for CMD58 (read OCR) */
#define SD_CARD_ERROR_CMD58         0x04
/** card returned an error response for CMD16 (set block len) */
#define SD_CARD_ERROR_CMD16         0x05
/** card operation voltage range doesn't match (2.7V - 3.6V) */
#define SD_CARD_ERROR_VOLTMATCH     0x06
/** timeout while waiting for start of read data */
#define SD_CARD_ERROR_READ_TIMEOUT  0x07
/** card returned error token when tried to read data */
#define SD_CARD_ERROR_READ          0x08
/** card returned an error response for CMD17 (read single block) */
#define SD_CARD_ERROR_CMD17         0x09
/** card returned an error token when tried to write data */
#define SD_CARD_ERROR_WRITE         0x0a
/** card returned an error response for CMD24 (write single block) */
#define SD_CARD_ERROR_CMD24         0x0b
/** timeout while waiting for writing data */
#define SD_CARD_ERROR_WRITE_TIMEOUT 0x0c
/** card returned an error in status after a data writing command */
#define SD_CARD_ERROR_WRITE_PROGR   0x0d
/** card returned an error response for CMD9  (read CSD) */
#define SD_CARD_ERROR_CMD9          0x0e
/** card returned an error response for CMD10 (read CID) */
#define SD_CARD_ERROR_CMD10         0x0f
/** card returned an error response for CMD18 (read multi block) */
#define SD_CARD_ERROR_CMD18         0x10
/** card returned an error response for CMD12 (stop block operation) */
#define SD_CARD_ERROR_CMD12         0x11
/** card returned an error response for ACMD23 (set pre-erase count) */
#define SD_CARD_ERROR_ACMD23        0x12
/** card returned an error response for CMD25 (write multi block) */
#define SD_CARD_ERROR_CMD25         0x13
/** card did not become ready for a multiple block write */
#define SD_CARD_ERROR_WRITE_MB      0x14 
/** timeout after sending STOP_TRAN_TOKEN after write multi block data */
#define SD_CARD_ERROR_STOP_TRAN     0x15

// card types
/** Standard capacity V1 SD card */
#define SD_CARD_TYPE_SD1  1
/** Standard capacity V2 SD card */
#define SD_CARD_TYPE_SD2  2
/** High Capacity SD card */
#define SD_CARD_TYPE_SDHC 3

#endif




//##########################################################################################################################################################




#ifndef SD_L2_H
#define SD_L2_H

#define SD_ENABLE_DIR_VIEW          1

/** No valid MBR/FAT-BS signature found in sector 0 */
#define SD_L2_ERROR_INVAL_SECT0     0x30
/** Malformed FAT boot sector */
#define SD_L2_ERROR_INVAL_BS        0x31
/** FAT12 is not supported */
#define SD_L2_ERROR_FAT12           0x32
/** FAT not initialized properly */
#define SD_L2_ERROR_FAT_NOT_INIT    0x33
/** End of cluster reached (not a real error, just information) */
#define SD_L2_ERROR_DIR_EOC         0x34
/** File not found after reaching end of directory */
#define SD_L2_ERROR_FILE_NOT_FOUND  0x35
/** Fragmentation found */
#define SD_L2_ERROR_FRAGMET_FOUND   0x36
/** Error in FAT entry */
#define SD_L2_ERROR_FAT_ENTRY       0x37
/** End of file reached */
#define SD_L2_ERROR_EOF             0x38
/** WorkBuf not set */
#define SD_L2_ERROR_WORKBUF         0x39


#define SD_L2_PARTTYPE_UNKNOWN      0
#define SD_L2_PARTTYPE_SUPERFLOPPY  1
#define SD_L2_PARTTYPE_FAT16        2
#define SD_L2_PARTTYPE_FAT32        3

typedef struct  {
    uint8_t     PartType;       // Use this to test whether it is FAT16 or FAT32 or not initialized
  
  // Stuff from FAT boot sector
  uint8_t     SecPerClus;
  uint16_t    RsvdSecCnt;
  uint8_t     NumFATs;
  uint16_t    RootEntryCount;
  uint32_t    TotalSec;
  uint32_t    SecPerFAT;
  uint32_t    RootClus;
  
  // For cluster calculations
  uint8_t     ClusterSizeShift;
  uint32_t    ClusterCount;
  
  // Start addresses (all in blocks / sector addresses)
  uint32_t    BootSectorStart;    // Address of boot sector from FAT
  uint32_t    FatStart;           // First file allocation table starts here
  uint32_t    RootDirStart;       // Root directory starts here
  uint32_t    DataStart;          // Cluster 0 starts here
  
  uint32_t    ClusterEndMarker;   // if Cluster >= this then end of file reached.
  
} SD_L2_FAT_t;

typedef struct {
  uint8_t     Attributes;
  uint32_t    Size;           // in bytes
  uint32_t    FirstCluster;   // First cluster
    
  uint32_t    ActSector;      // 0 to (SD_L2_FAT.SecPerClus - 1)
  uint32_t    ActBytePos;     // 0 to Size
} SD_L2_File_t;

extern SD_L2_FAT_t SD_L2_FAT;

uint8_t     SD_L2_Init(uint8_t *pWorkBuf);
void        SD_L2_DeInit();

uint8_t     SD_L2_SearchFile(uint8_t *filename, const uint32_t cluster, const uint8_t maskSet, const uint8_t maskUnset, SD_L2_File_t *fileinfo);
uint32_t    SD_L2_Cluster2Sector(uint32_t cluster);

uint8_t     SD_L2_IsFileFragmented(SD_L2_File_t *fileinfo);

#if SD_ENABLE_DIR_VIEW 
  uint8_t SD_L2_Dir(const uint32_t cluster, const uint8_t maskSet, const uint8_t maskUnset, void (*callback)(char *entryLine));
#endif  /* SD_ENABLE_DIR_VIEW */

#endif

